@@ -1,3 +1,16 @@
+// Copyright 2020-2022 The NATS Authors
+// Licensed under the Apache License, Version 2.0 (the "License");
+// you may not use this file except in compliance with the License.
+// You may obtain a copy of the License at
+//
+// http://www.apache.org/licenses/LICENSE-2.0
+//
+// Unless required by applicable law or agreed to in writing, software
+// distributed under the License is distributed on an "AS IS" BASIS,
+// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
+// See the License for the specific language governing permissions and
+// limitations under the License.
+
 //! An async Rust client for the NATS.io ecosystem.
 //!
 //! `git clone https://github.com/nats-io/nats.rs`
@@ -98,7 +111,8 @@ use std::time::Duration;
 use blocking::unblock;
 use crossbeam_channel::{Receiver, Sender};
 
-use crate::Headers;
+use crate::header::HeaderMap;
+use crate::IntoServerList;
 
 /// Connect to a NATS server at the given url.
 ///
@@ -108,7 +122,10 @@ use crate::Headers;
 /// let nc = nats::asynk::connect("demo.nats.io").await?;
 /// # std::io::Result::Ok(()) });
 /// ```
-pub async fn connect(nats_url: &str) -> io::Result<Connection> {
+pub async fn connect<I>(nats_url: I) -> io::Result<Connection>
+where
+    I: IntoServerList,
+{
     Options::new().connect(nats_url).await
 }
 
@@ -275,7 +292,7 @@ impl Connection {
         &self,
         subject: &str,
         reply: Option<&str>,
-        headers: Option<&Headers>,
+        headers: Option<&HeaderMap>,
         msg: impl AsRef<[u8]>,
     ) -> io::Result<()> {
         if let Some(res) = self
@@ -286,7 +303,7 @@ impl Connection {
         }
         let subject = subject.to_string();
         let reply = reply.map(str::to_owned);
-        let headers = headers.map(Headers::clone);
+        let headers = headers.map(HeaderMap::clone);
         let msg = msg.as_ref().to_vec();
         let inner = self.inner.clone();
         unblock(move || {
@@ -376,11 +393,11 @@ pub struct Message {
     pub data: Vec<u8>,
 
     /// Optional headers associated with this `Message`.
-    pub headers: Option<Headers>,
+    pub headers: Option<HeaderMap>,
 
     /// Client for publishing on the reply subject.
     #[doc(hidden)]
-    pub client: crate::Client,
+    pub client: Option<crate::client::Client>,
 
     /// Whether this message has already been successfully double-acked
     /// using `JetStream`.
@@ -402,22 +419,56 @@ impl From<crate::Message> for Message {
 }
 
 impl Message {
+    /// Creates new empty `Message`, without a Client.
+    /// Useful for passing `Message` data or creating `Message` instance without caring about `Client`,
+    /// but cannot be used on it's own for associated methods as those require `Client` injected into `Message`
+    /// and will error without it.
+    pub fn new(
+        subject: &str,
+        reply: Option<&str>,
+        data: impl AsRef<[u8]>,
+        headers: Option<HeaderMap>,
+    ) -> Message {
+        Message {
+            subject: subject.to_string(),
+            reply: reply.map(String::from),
+            data: data.as_ref().to_vec(),
+            headers,
+            ..Default::default()
+        }
+    }
+
     /// Respond to a request message.
     pub async fn respond(&self, msg: impl AsRef<[u8]>) -> io::Result<()> {
-        match self.reply.as_ref() {
-            None => Err(io::Error::new(
-                io::ErrorKind::InvalidInput,
-                "no reply subject available",
-            )),
-            Some(reply) => {
-                if let Some(res) = self.client.try_publish(reply, None, None, msg.as_ref()) {
-                    return res;
-                }
-                let reply = reply.to_string();
-                let msg = msg.as_ref().to_vec();
-                let client = self.client.clone();
-                unblock(move || client.publish(&reply, None, None, msg.as_ref())).await
-            }
+        let reply = self.reply.as_ref().ok_or_else(|| {
+            io::Error::new(io::ErrorKind::InvalidInput, "No reply subject to reply to")
+        })?;
+        let client = self.client.as_ref().ok_or_else(|| {
+            io::Error::new(
+                io::ErrorKind::NotConnected,
+                crate::message::MESSAGE_NOT_BOUND,
+            )
+        })?;
+        if let Some(res) = client.try_publish(reply.as_str(), None, None, msg.as_ref()) {
+            return res;
+        }
+        // clone only if we have to move the data to the thread
+        let client = client.clone();
+        let reply = reply.to_owned();
+        let msg = msg.as_ref().to_vec();
+        unblock(move || client.publish(&reply, None, None, msg.as_ref())).await
+    }
+}
+
+impl Default for Message {
+    fn default() -> Message {
+        Message {
+            subject: String::from(""),
+            reply: None,
+            data: Vec::new(),
+            headers: None,
+            client: None,
+            double_acked: Arc::new(AtomicBool::new(false)),
         }
     }
 }
@@ -605,6 +656,28 @@ impl Options {
         }
     }
 
+    /// Select option to enable reconnect with backoff
+    /// on first failed connection attempt.
+    /// The reconnect logic with `max_reconnects` and the
+    /// `reconnect_delay_callback` will be specified the same
+    /// as before but will be invoked on the first failed
+    /// connection attempt.
+    ///
+    /// # Example
+    /// ```
+    /// # smol::block_on(async {
+    /// let nc = nats::asynk::Options::new()
+    ///     .retry_on_failed_connect()
+    ///     .connect("demo.nats.io")
+    ///     .await?;
+    /// # std::io::Result::Ok(()) });
+    /// ```
+    pub fn retry_on_failed_connect(self) -> Options {
+        Options {
+            inner: self.inner.retry_on_failed_connect(),
+        }
+    }
+
     /// Set the maximum number of reconnect attempts.
     /// If no servers remain that are under this threshold,
     /// then no further reconnect shall be attempted.
@@ -677,9 +750,12 @@ impl Options {
     ///     .await?;
     /// # std::io::Result::Ok(()) });
     /// ```
-    pub async fn connect(self, nats_url: &str) -> io::Result<Connection> {
-        let nats_url = nats_url.to_string();
-        let conn = unblock(move || self.inner.connect(&nats_url)).await?;
+    pub async fn connect<I>(self, nats_url: I) -> io::Result<Connection>
+    where
+        I: IntoServerList,
+    {
+        let servers = nats_url.into_server_list()?;
+        let conn = unblock(move || self.inner.connect(servers)).await?;
         Ok(Connection::new(conn))
     }
 
