@@ -1,4 +1,4 @@
-// Copyright 2020-2021 The NATS Authors
+// Copyright 2020-2022 The NATS Authors
 // Licensed under the Apache License, Version 2.0 (the "License");
 // you may not use this file except in compliance with the License.
 // You may obtain a copy of the License at
@@ -11,7 +11,7 @@
 // See the License for the specific language governing permissions and
 // limitations under the License.
 
-use std::collections::{HashMap, VecDeque};
+use std::collections::{HashMap, HashSet, VecDeque};
 use std::fmt;
 use std::io::prelude::*;
 use std::io::{self, BufReader, BufWriter, Error, ErrorKind};
@@ -25,10 +25,10 @@ use crossbeam_channel as channel;
 use crossbeam_channel::RecvTimeoutError;
 use parking_lot::Mutex;
 
-use crate::connector::{Connector, NatsStream};
+use crate::connector::{Connector, NatsStream, ServerAddress};
 use crate::message::Message;
 use crate::proto::{self, ClientOp, ServerOp};
-use crate::{inject_delay, inject_io_failure, Headers, Options, ServerInfo};
+use crate::{header::HeaderMap, inject_delay, inject_io_failure, Options, ServerInfo};
 
 const BUF_CAPACITY: usize = 32 * 1024;
 
@@ -41,6 +41,12 @@ const BUF_CAPACITY: usize = 32 * 1024;
 struct State {
     write: Mutex<WriteState>,
     read: Mutex<ReadState>,
+    meta: Mutex<MetaState>,
+}
+
+struct MetaState {
+    /// Set of subjects that are currently muted.
+    mutes: HashSet<u64>,
 }
 
 struct WriteState {
@@ -77,11 +83,15 @@ struct ReadState {
     pings_out: u8,
 }
 
+/// A predicate used to preprocess messages for a subscription as they arrive over the wire.
+pub(crate) type Preprocessor = Box<dyn Fn(u64, &Message) -> bool + Send + Sync>;
+
 /// A registered subscription.
 struct Subscription {
     subject: String,
     queue_group: Option<String>,
     messages: channel::Sender<Message>,
+    preprocess: Preprocessor,
 }
 
 /// A NATS client.
@@ -102,7 +112,7 @@ pub struct Client {
 
 impl Client {
     /// Creates a new client that will begin connecting in the background.
-    pub(crate) fn connect(url: &str, options: Options) -> io::Result<Client> {
+    pub(crate) fn connect(urls: Vec<ServerAddress>, options: Options) -> io::Result<Client> {
         // A channel for coordinating flushes.
         let (flush_kicker, flush_wanted) = channel::bounded(1);
 
@@ -113,6 +123,9 @@ impl Client {
         // The client state.
         let client = Client {
             state: Arc::new(State {
+                meta: Mutex::new(MetaState {
+                    mutes: HashSet::new(),
+                }),
                 write: Mutex::new(WriteState {
                     writer: None,
                     flush_kicker,
@@ -135,7 +148,7 @@ impl Client {
 
         // Connector for creating the initial connection and reconnecting when
         // it is broken.
-        let connector = Connector::new(url, options.clone())?;
+        let connector = Connector::new(urls, options.clone())?;
 
         // Spawn the client thread responsible for:
         // - Maintaining a connection to the server and reconnecting when it is
@@ -195,18 +208,21 @@ impl Client {
 
                             // Flush the writer.
                             let mut write = client.state.write.lock();
+                            let mut read = client.state.read.lock();
+
                             if let Some(writer) = write.writer.as_mut() {
                                 let res = writer.flush();
                                 last = Instant::now();
                                 // If flushing fails, disconnect.
                                 if res.is_err() {
-                                    // NB see locking protocol for state.write and state.read
                                     writer.get_ref().shutdown();
                                     write.writer = None;
-                                    let mut read = client.state.read.lock();
                                     read.pongs.clear();
                                 }
                             }
+
+                            // NB see locking protocol for state.write and state.read
+                            drop(read);
                             drop(write);
                         }
                         Err(RecvTimeoutError::Timeout) => {
@@ -236,6 +252,7 @@ impl Client {
                                 }
                             }
 
+                            // NB see locking protocol for state.write and state.read
                             drop(read);
                             drop(write);
                         }
@@ -360,7 +377,16 @@ impl Client {
         subject: &str,
         queue_group: Option<&str>,
     ) -> io::Result<(u64, channel::Receiver<Message>)> {
-        // Inject random delays when testing.
+        self.subscribe_with_preprocessor(subject, queue_group, Box::new(|_, _| false))
+    }
+
+    /// Subscribe to a subject with a message preprocessor.
+    pub(crate) fn subscribe_with_preprocessor(
+        &self,
+        subject: &str,
+        queue_group: Option<&str>,
+        message_processor: Preprocessor,
+    ) -> io::Result<(u64, channel::Receiver<Message>)> {
         inject_delay();
 
         let mut write = self.state.write.lock();
@@ -392,6 +418,7 @@ impl Client {
                 subject: subject.to_string(),
                 queue_group: queue_group.map(ToString::to_string),
                 messages: sender,
+                preprocess: message_processor,
             },
         );
 
@@ -402,6 +429,66 @@ impl Client {
         Ok((sid, receiver))
     }
 
+    /// Marks a subscription as muted.
+    pub(crate) fn mute(&self, sid: u64) -> io::Result<bool> {
+        let mut meta = self.state.meta.lock();
+        Ok(meta.mutes.insert(sid))
+    }
+
+    /// Resubscribes an existing subscription by unsubscribing from the old subject and subscribing
+    /// to the new subject returning a new sid while retaining the existing channel receiver.
+    pub(crate) fn resubscribe(&self, old_sid: u64, new_subject: &str) -> io::Result<u64> {
+        // Inject random delays when testing.
+        inject_delay();
+
+        let mut write = self.state.write.lock();
+        let mut read = self.state.read.lock();
+
+        // Check if the client is closed.
+        self.check_shutdown()?;
+
+        let subscription = read
+            .subscriptions
+            .remove(&old_sid)
+            .ok_or_else(|| io::Error::new(io::ErrorKind::InvalidInput, "subscription not found"))?;
+
+        // Generate a subject ID.
+        let new_sid = write.next_sid;
+        write.next_sid += 1;
+
+        // Send an UNSUB and SUB messages.
+        if let Some(writer) = write.writer.as_mut() {
+            proto::encode(
+                writer,
+                ClientOp::Unsub {
+                    sid: old_sid,
+                    max_msgs: None,
+                },
+            )?;
+        }
+
+        let queue_group = subscription.queue_group.clone();
+        read.subscriptions.insert(new_sid, subscription);
+
+        if let Some(writer) = write.writer.as_mut() {
+            proto::encode(
+                writer,
+                ClientOp::Sub {
+                    sid: new_sid,
+                    subject: new_subject,
+                    queue_group: queue_group.as_deref(),
+                },
+            )?;
+            write.flush_kicker.try_send(()).ok();
+        }
+
+        // NB see locking protocol for state.write and state.read
+        drop(read);
+        drop(write);
+
+        Ok(new_sid)
+    }
+
     /// Unsubscribes from a subject.
     pub(crate) fn unsubscribe(&self, sid: u64) -> io::Result<()> {
         // Inject random delays when testing.
@@ -440,7 +527,7 @@ impl Client {
         &self,
         subject: &str,
         reply_to: Option<&str>,
-        headers: Option<&Headers>,
+        headers: Option<&HeaderMap>,
         msg: &[u8],
     ) -> io::Result<()> {
         // Inject random delays when testing.
@@ -514,7 +601,7 @@ impl Client {
         &self,
         subject: &str,
         reply_to: Option<&str>,
-        headers: Option<&Headers>,
+        headers: Option<&HeaderMap>,
         msg: &[u8],
     ) -> Option<io::Result<()>> {
         // Check if the client is closed.
@@ -586,10 +673,12 @@ impl Client {
         let mut first_connect = true;
 
         loop {
-            // Don't use backoff on first connect.
-            let use_backoff = !first_connect;
+            //  Don't use backoff on first connect unless retry_on_failed_connect is set to true.
+            let use_backoff = self.options.retry_on_failed_connect || !first_connect;
+
             // Make a connection to the server.
             let (server_info, stream) = connector.connect(use_backoff)?;
+            self.process_info(&server_info, &connector);
 
             let reader = BufReader::with_capacity(BUF_CAPACITY, stream.clone());
             let writer = BufWriter::with_capacity(BUF_CAPACITY, stream);
@@ -685,6 +774,13 @@ impl Client {
         Ok(())
     }
 
+    // processes action need to be performed based on retrieved server info.
+    fn process_info(&self, server_info: &ServerInfo, connector: &Connector) {
+        if server_info.lame_duck_mode {
+            connector.get_options().lame_duck_callback.call();
+        }
+    }
+
     /// Updates our last activity from the server.
     fn update_activity(&self) {
         let mut read = self.state.read.lock();
@@ -708,8 +804,9 @@ impl Client {
             match op {
                 ServerOp::Info(server_info) => {
                     for url in &server_info.connect_urls {
-                        connector.add_url(url).ok();
+                        connector.add_server(url.parse()?);
                     }
+                    self.process_info(&server_info, connector);
                     *self.server_info.lock() = server_info;
                 }
 
@@ -757,6 +854,11 @@ impl Client {
                     reply_to,
                     payload,
                 } => {
+                    // Ignore muted subscriptions
+                    if self.state.meta.lock().mutes.get(&sid).is_some() {
+                        continue;
+                    }
+
                     let read = self.state.read.lock();
 
                     // Send the message to matching subscription.
@@ -766,13 +868,20 @@ impl Client {
                             reply: reply_to,
                             data: payload,
                             headers: None,
-                            client: self.clone(),
+                            client: Some(self.clone()),
                             double_acked: Default::default(),
                         };
 
+                        // Preprocess and drop the message from the buffer if it the predicate
+                        // returns true
+                        let preprocess = &subscription.preprocess;
+                        if preprocess(sid, &msg) {
+                            continue;
+                        }
+
                         // Send a message or drop it if the channel is
                         // disconnected or full.
-                        subscription.messages.try_send(msg).ok();
+                        subscription.messages.send(msg).unwrap();
                     }
                 }
 
@@ -783,6 +892,11 @@ impl Client {
                     reply_to,
                     payload,
                 } => {
+                    // Ignore muted subscriptions
+                    if self.state.meta.lock().mutes.get(&sid).is_some() {
+                        continue;
+                    }
+
                     let read = self.state.read.lock();
                     // Send the message to matching subscription.
                     if let Some(subscription) = read.subscriptions.get(&sid) {
@@ -791,18 +905,28 @@ impl Client {
                             reply: reply_to,
                             data: payload,
                             headers: Some(headers),
-                            client: self.clone(),
+                            client: Some(self.clone()),
                             double_acked: Default::default(),
                         };
 
+                        // Preprocess and drop the message from the buffer if it the predicate
+                        // returns true
+                        let preprocess = &subscription.preprocess;
+                        if preprocess(sid, &msg) {
+                            continue;
+                        }
+
                         // Send a message or drop it if the channel is
                         // disconnected or full.
-                        subscription.messages.try_send(msg).ok();
+                        subscription.messages.send(msg).unwrap();
                     }
                 }
 
                 ServerOp::Err(msg) => {
-                    return Err(Error::new(ErrorKind::Other, msg));
+                    connector
+                        .get_options()
+                        .error_callback
+                        .call(self, Error::new(ErrorKind::Other, msg));
                 }
 
                 ServerOp::Unknown(line) => {
