@@ -1,4 +1,4 @@
-// Copyright 2020-2021 The NATS Authors
+// Copyright 2020-2022 The NATS Authors
 // Licensed under the Apache License, Version 2.0 (the "License");
 // you may not use this file except in compliance with the License.
 // You may obtain a copy of the License at
@@ -115,6 +115,7 @@
 //! # Ok(()) }
 //! ```
 
+#![cfg_attr(docsrs, feature(doc_cfg))]
 #![cfg_attr(test, deny(warnings))]
 #![cfg_attr(
     feature = "fault_injection",
@@ -164,16 +165,12 @@
     clippy::match_same_arms,
     clippy::maybe_infinite_iter,
     clippy::mem_forget,
-    clippy::module_name_repetitions,
     clippy::needless_borrow,
     clippy::needless_continue,
     clippy::needless_pass_by_value,
     clippy::non_ascii_literal,
     clippy::path_buf_push_overwrite,
     clippy::print_stdout,
-    clippy::shadow_reuse,
-    clippy::shadow_same,
-    clippy::shadow_unrelated,
     clippy::single_match_else,
     clippy::string_add,
     clippy::string_add_assign,
@@ -188,7 +185,10 @@
     clippy::match_like_matches_macro,
     clippy::await_holding_lock,
     clippy::shadow_reuse,
-    clippy::wildcard_enum_match_arm
+    clippy::shadow_same,
+    clippy::shadow_unrelated,
+    clippy::wildcard_enum_match_arm,
+    clippy::module_name_repetitions
 )]
 
 /// Async-enabled NATS client.
@@ -198,17 +198,24 @@ mod auth_utils;
 mod client;
 mod connect;
 mod connector;
-mod headers;
-mod jetstream_types;
 mod message;
 mod options;
 mod proto;
 mod secure_wipe;
 mod subscription;
 
+/// Header constants and types.
+pub mod header;
+
 /// `JetStream` stream management and consumers.
 pub mod jetstream;
 
+#[cfg_attr(docsrs, doc(cfg(feature = "unstable")))]
+pub mod kv;
+
+#[cfg_attr(docsrs, doc(cfg(feature = "unstable")))]
+pub mod object_store;
+
 #[cfg(feature = "fault_injection")]
 mod fault_injection;
 
@@ -223,20 +230,33 @@ fn inject_io_failure() -> io::Result<()> {
     Ok(())
 }
 
+// comment out until we reach MSRV 1.54.0
+// #[doc = include_str!("../docs/migration-guide-0.17.0.md")]
+// #[derive(Copy, Clone)]
+// pub struct Migration0170;
+
 #[doc(hidden)]
 #[deprecated(since = "0.6.0", note = "this has been renamed to `Options`.")]
 pub type ConnectionOptions = Options;
 
+#[doc(hidden)]
+#[deprecated(since = "0.17.0", note = "this has been moved to `header::HeaderMap`.")]
+pub type Headers = HeaderMap;
+
 use std::{
     io::{self, Error, ErrorKind},
     sync::Arc,
     time::{Duration, Instant},
 };
 
-pub use headers::Headers;
+use lazy_static::lazy_static;
+use regex::Regex;
+
+pub use connector::{IntoServerList, ServerAddress};
+pub use jetstream::JetStreamOptions;
 pub use message::Message;
 pub use options::Options;
-pub use subscription::Subscription;
+pub use subscription::{Handler, Subscription};
 
 /// A re-export of the `rustls` crate used in this crate,
 /// for use in cases where manual client configurations
@@ -247,6 +267,7 @@ pub use rustls;
 pub use connect::ConnectInfo;
 
 use client::Client;
+use header::HeaderMap;
 use options::AuthStyle;
 use secure_wipe::{SecureString, SecureVec};
 
@@ -254,6 +275,10 @@ const VERSION: &str = env!("CARGO_PKG_VERSION");
 const LANG: &str = "rust";
 const DEFAULT_FLUSH_TIMEOUT: Duration = Duration::from_secs(10);
 
+lazy_static! {
+    static ref VERSION_RE: Regex = Regex::new(r#"\Av?([0-9]+)\.?([0-9]+)?\.?([0-9]+)?"#).unwrap();
+}
+
 /// Information sent by the server back to this client
 /// during initial connection, and possibly again later.
 #[allow(unused)]
@@ -290,6 +315,8 @@ pub struct ServerInfo {
     pub client_ip: String,
     /// Whether the server supports headers.
     pub headers: bool,
+    /// Whether server goes into lame duck mode.
+    pub lame_duck_mode: bool,
 }
 
 impl ServerInfo {
@@ -314,13 +341,14 @@ impl ServerInfo {
                 .collect(),
             client_ip: obj["client_ip"].take_string().unwrap_or_default(),
             headers: obj["headers"].as_bool().unwrap_or(false),
+            lame_duck_mode: obj["ldm"].as_bool().unwrap_or(false),
         })
     }
 }
 
 /// A NATS connection.
 #[derive(Clone, Debug)]
-pub struct Connection(Arc<Inner>);
+pub struct Connection(pub(crate) Arc<Inner>);
 
 #[derive(Clone, Debug)]
 struct Inner {
@@ -333,23 +361,68 @@ impl Drop for Inner {
     }
 }
 
-/// Connect to a NATS server at the given url.
+/// Connect to one or more NATS servers at the given URLs.
 ///
-/// # Example
+/// The [`IntoServerList`] trait allows to pass URLs in various different formats. Furthermore, if
+/// you need more control of the connection's parameters use [`Options::connect()`].
+///
+/// **Warning:** There are asynchronous errors that can happen during operation of NATS client.
+/// To handle them, add handler for [`Options::error_callback()`].
+///
+/// # Examples
+///
+/// If no scheme is provided the `nats://` scheme is assumed. The default port is `4222`.
 /// ```
-/// # fn main() -> std::io::Result<()> {
 /// let nc = nats::connect("demo.nats.io")?;
-/// # Ok(())
-/// # }
+/// # Ok::<(), std::io::Error>(())
 /// ```
-pub fn connect(nats_url: &str) -> io::Result<Connection> {
-    Options::new().connect(nats_url)
+///
+/// It is possible to provide several URLs as a comma separated list.
+/// ```
+/// let nc = nats::connect("demo.nats.io,tls://demo.nats.io:4443")?;
+/// # Ok::<(), std::io::Error>(())
+/// ```
+///
+/// Alternatively, an array of strings can be passed.
+/// ```
+/// # use nats::IntoServerList;
+/// let nc = nats::connect(&["demo.nats.io", "tls://demo.nats.io:4443"])?;
+/// # Ok::<(), std::io::Error>(())
+/// ```
+///
+/// Instead of using strings, [`ServerAddress`]es can be used directly as well. This is handy for
+/// validating user input.
+/// ```
+/// use std::io;
+/// use structopt::StructOpt;
+/// use nats::ServerAddress;
+///
+/// #[derive(Debug, StructOpt)]
+/// struct Config {
+///     #[structopt(short, long = "server", default_value = "demo.nats.io")]
+///     servers: Vec<ServerAddress>,
+/// }
+///
+/// fn main() -> io::Result<()> {
+///     let config = Config::from_args();
+///     let nc = nats::connect(config.servers)?;
+///     Ok(())
+/// }
+/// ```
+pub fn connect<I: IntoServerList>(nats_urls: I) -> io::Result<Connection> {
+    Options::new().connect(nats_urls)
 }
 
 impl Connection {
-    /// Connects on a URL with the given options.
-    pub(crate) fn connect_with_options(url: &str, options: Options) -> io::Result<Connection> {
-        let client = Client::connect(url, options)?;
+    /// Connects on one or more NATS servers with the given options.
+    ///
+    /// For more on how to use [`IntoServerList`] trait see [`crate::connect()`].
+    pub(crate) fn connect_with_options<I>(urls: I, options: Options) -> io::Result<Connection>
+    where
+        I: IntoServerList,
+    {
+        let urls = urls.into_server_list()?;
+        let client = Client::connect(urls, options)?;
         client.flush(DEFAULT_FLUSH_TIMEOUT)?;
         Ok(Connection(Arc::new(Inner { client })))
     }
@@ -448,20 +521,7 @@ impl Connection {
     /// # }
     /// ```
     pub fn request(&self, subject: &str, msg: impl AsRef<[u8]>) -> io::Result<Message> {
-        // Publish a request.
-        let reply = self.new_inbox();
-        let sub = self.subscribe(&reply)?;
-        self.publish_with_reply_or_headers(subject, Some(reply.as_str()), None, msg)?;
-
-        // Wait for the response.
-        if let Some(msg) = sub.next() {
-            if msg.is_no_responders() {
-                return Err(Error::new(ErrorKind::NotFound, "no responders"));
-            }
-            Ok(msg)
-        } else {
-            Err(ErrorKind::ConnectionReset.into())
-        }
+        self.request_with_headers_or_timeout(subject, None, None, msg)
     }
 
     /// Publish a message on the given subject as a request and receive the
@@ -482,14 +542,39 @@ impl Connection {
         subject: &str,
         msg: impl AsRef<[u8]>,
         timeout: Duration,
+    ) -> io::Result<Message> {
+        self.request_with_headers_or_timeout(subject, None, Some(timeout), msg)
+    }
+
+    fn request_with_headers_or_timeout(
+        &self,
+        subject: &str,
+        maybe_headers: Option<&HeaderMap>,
+        maybe_timeout: Option<Duration>,
+        msg: impl AsRef<[u8]>,
     ) -> io::Result<Message> {
         // Publish a request.
         let reply = self.new_inbox();
         let sub = self.subscribe(&reply)?;
-        self.publish_with_reply_or_headers(subject, Some(reply.as_str()), None, msg)?;
+        self.publish_with_reply_or_headers(subject, Some(reply.as_str()), maybe_headers, msg)?;
 
-        // Wait for the response.
-        sub.next_timeout(timeout)
+        // Wait for the response
+        let result = if let Some(timeout) = maybe_timeout {
+            sub.next_timeout(timeout)
+        } else if let Some(msg) = sub.next() {
+            Ok(msg)
+        } else {
+            Err(ErrorKind::ConnectionReset.into())
+        };
+
+        // Check for no responder status.
+        if let Ok(msg) = result.as_ref() {
+            if msg.is_no_responders() {
+                return Err(Error::new(ErrorKind::NotFound, "no responders"));
+            }
+        }
+
+        result
     }
 
     /// Publish a message on the given subject as a request and allow multiple
@@ -590,6 +675,35 @@ impl Connection {
         Ok(start.elapsed())
     }
 
+    /// Returns true if the version is compatible with the version components.
+    pub fn is_server_compatible_version(&self, major: i64, minor: i64, patch: i64) -> bool {
+        let server_info = self.0.client.server_info();
+        let server_version_captures = VERSION_RE.captures(&server_info.version).unwrap();
+        let server_major = server_version_captures
+            .get(1)
+            .map(|m| m.as_str().parse::<i64>().unwrap())
+            .unwrap();
+
+        let server_minor = server_version_captures
+            .get(2)
+            .map(|m| m.as_str().parse::<i64>().unwrap())
+            .unwrap();
+
+        let server_patch = server_version_captures
+            .get(3)
+            .map(|m| m.as_str().parse::<i64>().unwrap())
+            .unwrap();
+
+        if server_major < major
+            || (server_major == major && server_minor < minor)
+            || (server_major == major && server_minor == minor && server_patch < patch)
+        {
+            return false;
+        }
+
+        true
+    }
+
     /// Returns the client IP as known by the server.
     /// Supported as of server version 2.1.6.
     /// # Example
@@ -703,7 +817,7 @@ impl Connection {
         &self,
         subject: &str,
         reply: Option<&str>,
-        headers: Option<&Headers>,
+        headers: Option<&HeaderMap>,
         msg: impl AsRef<[u8]>,
     ) -> io::Result<()> {
         self.0.client.publish(subject, reply, headers, msg.as_ref())
@@ -739,7 +853,7 @@ impl Connection {
         &self,
         subject: &str,
         reply: Option<&str>,
-        headers: Option<&Headers>,
+        headers: Option<&HeaderMap>,
         msg: impl AsRef<[u8]>,
     ) -> Option<io::Result<()>> {
         self.0
