@@ -1,3 +1,16 @@
+// Copyright 2020-2022 The NATS Authors
+// Licensed under the Apache License, Version 2.0 (the "License");
+// you may not use this file except in compliance with the License.
+// You may obtain a copy of the License at
+//
+// http://www.apache.org/licenses/LICENSE-2.0
+//
+// Unless required by applicable law or agreed to in writing, software
+// distributed under the License is distributed on an "AS IS" BASIS,
+// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
+// See the License for the specific language governing permissions and
+// limitations under the License.
+
 use smol::future::FutureExt;
 use std::io;
 
@@ -44,7 +57,7 @@ fn async_subscription_drop() -> io::Result<()> {
     let s = util::run_basic_server();
 
     smol::block_on(async {
-        let nc = nats::asynk::connect(&s.client_url()).await?;
+        let nc = nats::asynk::connect(s.client_url()).await?;
 
         let inbox = nc.new_inbox();
 
